/****************************************************************************/
/** G-FIT V1.0                                                             **/
/** Fitting hyper-Erlang distributions to trace data with the EM algorithm **/
/**                                                                        **/
/** Developed by Axel Thuemmler, July/August 2004                          **/
/** Copyright (c) 2004 University of Dortmund. All rights reserved.        **/
/**                                                                        **/
/** The G-FIT software package is distributed in open-source under the     **/
/** terms of the GNU general public license agreement. To document the     **/
/** usage of G-FIT, we ask each user to cite our scientific publications   **/
/** mentioned on our Web site.                                             **/
/****************************************************************************/
//------------------------------------------------------
// Methods for evaluating the hyper-Erlang distribution
//------------------------------------------------------

#include <iostream>
#include <cstdlib>
#include <fstream>
#include <cmath>

#include "gerlangdist.h"


using namespace std;


TGerlangDist::TGerlangDist(unsigned int _m)
{
	m = _m;
	r = new unsigned int[m];
	lambdas = new double[m];
	weights = new double[m];

	// initializes pre-computed faculty values
	init_faculties();
}


TGerlangDist::~TGerlangDist()
{
	delete [] r;
	delete [] lambdas;
	delete [] weights;
}


//-------------------------------------------------
// Initialize hyper-Erlang distribution parameters
//-------------------------------------------------
void TGerlangDist::init(unsigned int* _r, double* _lambdas, double* _weights)
{
	for (unsigned int i = 0; i < m; i++)
	{
		r[i] = _r[i];

		if (_lambdas != NULL)
		{
			lambdas[i] = _lambdas[i];
		}
		else
		{
			lambdas[i] = (double) (i+1);
		}

		if (_weights != NULL)
		{
			weights[i] = _weights[i];
		}
		else
		{
			weights[i] = 1.0 / (double) m;
		}
	}
} 


//--------------------------------------------------------------------
// Copy hyper-Erlang distribution parameters in pre-allocated storage
//--------------------------------------------------------------------
void TGerlangDist::get(unsigned int* _r, double* _lambdas, double* _weights)
{
	for (unsigned int i = 0; i < m; i++)
	{
		_r[i] = r[i];
		_lambdas[i] = lambdas[i];
		_weights[i] = weights[i];
	}
}


//--------------------------------------------
// Adjust number of different Erlang branches
//--------------------------------------------
void TGerlangDist::adjust_m(unsigned int m_new)
{
	if (m != m_new)
	{
		delete [] r;
		delete [] lambdas;
		delete [] weights;

		m = m_new;
		r = new unsigned int[m];
		lambdas = new double[m];
		weights = new double[m];
	}
}


void TGerlangDist::print_r()
{
	for (unsigned int i = 0; i < m; i++)
	{
		cout << "r_" << i+1 << " = " << r[i] << endl;
	}
}

void TGerlangDist::print_lambdas()
{
	for (unsigned int i = 0; i < m; i++)
	{
		cout << "lambda_" << i+1 << " = " << lambdas[i] << endl;
	}
}

void TGerlangDist::print_weights()
{
	for (unsigned int i = 0; i < m; i++)
	{
		cout << "weight_" << i+1 << " = " << weights[i] << endl;
	}
}


//--------------------------------------------
// Initialize array with pre-computed faculty
// and log-faculty values
//--------------------------------------------
void TGerlangDist::init_faculties()
{
	fac[0] = 1;
	lnfac[0] = 0;
	for (unsigned int i = 1; i < MAX_FACULTY+1; i++)
	{
		fac[i] = i * fac[i-1];
		lnfac[i] = log( (double) i) + lnfac[i-1];
	}
}


//-------------------------------------------------------
// Compute Erlang probability density function (pdf)
// Fast and numerically stable implementation, that uses
// precomputed log-faculty values
//-------------------------------------------------------
double TGerlangDist::erlang_pdf(unsigned int r, double lambda, double x)
{
	if (x == 0)
	{
		if (r > 1) return 0.0;
		else return lambda;
	}
	else
	{
		return (lambda * exp((r-1) * log(lambda*x) - lnfac[r-1] - lambda * x));
	}
}


//-------------------------------------------------------
// Compute Erlang cumulative distribution function (cdf)
//-------------------------------------------------------
double TGerlangDist::erlang_cdf(unsigned int r, double lambda, double x)
{
	double sum = 1.0;
	double prod = 1.0;

	for (unsigned int i = 1; i < r; i++)
	{
		prod *= lambda * x / (double) i;
		sum += prod;
	}

	return 1 - exp(-lambda*x) * sum;
}


//---------------------------------------------------------
// Compute the log-likelihood function of the hyper-Erlang
// distribution for a given trace
//---------------------------------------------------------
double TGerlangDist::compute_loglikelihood(TTrace* trace)
{
	double pdf;

	double loglikelihood = 0.0;
	for (unsigned int n = 0; n < trace->size; n++)
	{
		pdf = 0.0;
		for (unsigned int i = 0; i < m; i++)
		{
			pdf += weights[i] * erlang_pdf(r[i], lambdas[i], trace->trace[n]);
		}
		loglikelihood += log(pdf);
	}

	return loglikelihood;
}


//-----------------------------------------------------------
// Compute pdf, cdf, and ccdf of hyper-Erlang distribution
//   width   = width of one step
//   steps   = number of steps to be evaluated
//             (pdf, cdf, and ccdf go from 0 to width*steps)
//-----------------------------------------------------------
void TGerlangDist::compute_distribution(const char* filename, double width, unsigned int steps)
{

	ofstream gfitfile(filename);
	double x, pdf, cdf;

	for (unsigned int n = 0; n < steps+1; n++)
	{
		x = n * width;

		// computation of hyper-Erlang pdf and cdf
		pdf = cdf = 0.0;
		for (unsigned int i = 0; i < m; i++)
		{
			pdf += weights[i] * erlang_pdf(r[i], lambdas[i], x);
			cdf += weights[i] * erlang_cdf(r[i], lambdas[i], x);
		}
		gfitfile << x << "  " << pdf << "  " << cdf << "  " << 1.0 - cdf << endl;
	}

	gfitfile.close();
}


//------------------------------------------------------------------------
// Compute first five moments, variance, coefficient of variation,
// minimum, and maximum of hyper-Erlang distribution
//   ts      = trace statistics needed for computation of relative errors
//------------------------------------------------------------------------
void TGerlangDist::compute_statistics(TTraceStatistics ts)
{

	double m1, m2, m3, m4, m5, var, cvs;
	m1 = m2 = m3 = m4 = m5 = var = 0.0;
	for (unsigned int i = 0; i < m; i++)
	{
		m1 += weights[i] * r[i] / lambdas[i];
		m2 += weights[i] * r[i] * (r[i]+1) / pow(lambdas[i],2);
		m3 += weights[i] * r[i] * (r[i]+1) * (r[i]+2) / pow(lambdas[i],3);
		m4 += weights[i] * r[i] * (r[i]+1) * (r[i]+2) * (r[i]+3) / pow(lambdas[i],4);
		m5 += weights[i] * r[i] * (r[i]+1) * (r[i]+2) * (r[i]+3) * (r[i]+4) / pow(lambdas[i],5);
	}
	var = m2 - m1*m1;
	cvs = var / (m1*m1);

	cout.setf(ios::floatfield);					// digits before and after decimal point
	cout.precision(10);							// are together maximal 10
	cout << "First   Moment         : " << m1;
	cout.setf(ios::fixed, ios::floatfield);		// exactly 2 digits after decimal point
	cout.precision(2);							// and arbitrary many before
	cout << " (" << fabs((m1-ts.m[0])/ts.m[0])*100 << "%)" << endl;
	cout.setf(ios::floatfield);
	cout.precision(10);
	cout << "Second  Moment         : " << m2;
	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(2);
	cout << " (" << fabs((m2-ts.m[1])/ts.m[1])*100 << "%)" << endl;
	cout.setf(ios::floatfield);
	cout.precision(10);
	cout << "Third   Moment         : " << m3;
	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(2);
	cout << " (" << fabs((m3-ts.m[2])/ts.m[2])*100 << "%)" << endl;
	cout.setf(ios::floatfield);
	cout.precision(10);
	cout << "Variance               : " << var;
	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(2);
	cout << " (" << fabs((var-ts.var)/ts.var)*100 << "%)" << endl;
	cout.setf(ios::floatfield);
	cout.precision(10);
	cout << "Squared Coeff. of Var. : " << cvs;
	cout.setf(ios::fixed, ios::floatfield);
	cout.precision(2);
	cout << " (" << fabs((cvs-ts.cvs)/ts.cvs)*100 << "%)" << endl;
	cout.setf(ios::floatfield);
	cout.precision(10);
}


//---------------------------------------
// Write distribution parameters to file
//---------------------------------------
void TGerlangDist::write_file(const char* filename)
{

	ofstream distfile(filename);
	distfile.precision(14);

	distfile << "# Parameters of hyper-Erlang distribution:" << endl;
	distfile << "# number of different Erlang branches, m" << endl;
	distfile << m << endl << endl;

	distfile << "# number of phases of i-th Erlang branch, i = 1,...,m" << endl;
	for (unsigned int i = 0; i < m; i++)
	{
		distfile << r[i] << endl;
	}
	distfile << endl;

	distfile << "# lambda values of i-th Erlang branch, i = 1,...,m" << endl;
	for (unsigned int i = 0; i < m; i++)
	{
		distfile << lambdas[i] << endl;
	}
	distfile << endl;

	distfile << "# initial probabilities of i-th Erlang branch, i = 1,...,m" << endl;
	for (unsigned int i = 0; i < m; i++)
	{
		distfile << weights[i] << endl;
	}
	distfile << endl;

	distfile.close();
}


//----------------------------------------
// Read distribution parameters form file
// and return number of Erlang branches
//----------------------------------------
unsigned int TGerlangDist::read_file(const char* filename)
{
	cerr << "reading hyper-Erlang distribution from file ...";

	ifstream gfitfile(filename);
	if (!gfitfile)
	{
		cerr << endl << "ERROR: file not found !!!" << endl;
		exit (-1);
	}

	bool m_read, r_read, l_read, w_read;
	m_read = r_read = l_read = w_read = false;
	unsigned int read_count = 0;

	char line[257];

	// get the hyper-Erlang distribution from G-FIT results file
	while (gfitfile.getline(line, 256))
	{
		// line empty => next line
		if (strcmp(line, "") == 0) continue;

		if (strcmp(line, "# number of different Erlang branches, m") == 0)
		{
			m_read = true;
			continue;
		}
		if (strcmp(line, "# number of phases of i-th Erlang branch, i = 1,...,m") == 0)
		{
			r_read = true;
			read_count = 0;
			continue;
		}
		if (strcmp(line, "# lambda values of i-th Erlang branch, i = 1,...,m") == 0)
		{
			l_read = true;
			read_count = 0;
			continue;
		}
		if (strcmp(line, "# initial probabilities of i-th Erlang branch, i = 1,...,m") == 0)
		{
			w_read = true;
			read_count = 0;
			continue;
		}

		// read parameter values
		if (m_read)
		{
			adjust_m(atol(line));
			m_read = false;
		}
		if (r_read)
		{
			r[read_count] = atol(line);
			read_count++;
			if (read_count == m) r_read = false;
		}
		if (l_read)
		{
			lambdas[read_count] = atof(line);
			read_count++;
			if (read_count == m) l_read = false;
		}
		if (w_read)
		{
			weights[read_count] = atof(line);
			read_count++;
			if (read_count == m) w_read = false;
		}
	}

	gfitfile.close();

	cout << endl;

	return m;
}
