/****************************************************************************/
/** G-FIT V1.0                                                             **/
/** Fitting hyper-Erlang distributions to trace data with the EM algorithm **/
/**                                                                        **/
/** Developed by Axel Thuemmler, July/August 2004                          **/
/** Copyright (c) 2004 University of Dortmund. All rights reserved.        **/
/**                                                                        **/
/** The G-FIT software package is distributed in open-source under the     **/
/** terms of the GNU general public license agreement. To document the     **/
/** usage of G-FIT, we ask each user to cite our scientific publications   **/
/** mentioned on our Web site.                                             **/
/****************************************************************************/

#include <iostream>
#include <sstream>
#include <ctime>

#include "trace.h"
#include "gerlangdist.h"
#include "gerlangfit.h"


//#define PLOT_WIDTH			0.15
//#define PLOT_WIDTH		0.066666666666666
#define PLOT_WIDTH		0.05
//#define PLOT_WIDTH		0.01

#define PLOT_STEPS		10000

//#define WRITE_SORTED_TRACE


// default values
#define DEFAULT_N		5
#define DEFAULT_M_MIN	1
#define DEFAULT_M_MAX	5

// accuracy of fitting procedure
#define DEFAULT_EPS 	1e-6
//#define DEFAULT_EPS		1e-3

// name of gnuplot data file for fitted HErD pdf, cdf, and ccdf
#define GFIT_FILENAME	"./curves/gfit.dat"

using namespace std;


enum TFitAlgorithm {FIT, FIT_ALL, FIT_ALL_STEPS, CALC_NUM};


//---------------------------------
// Print usage of the fitting tool
//---------------------------------
void usage(char **argv)
{
	cerr << "Usage: " << argv[0] << " <tracefile> [options]" << endl;
	cerr << "Explanation of options:" << endl;
	cerr << "  -f <m> <r_1> ... <r_m>  = fit a single hyper-Erlang distribution to the given data trace." << endl;
	cerr << "                            <m> is the number of different Erlang branches and" << endl;
	cerr << "                            <r_i> is the number of phases of the i-th Erlang branch (i = 1,...,m)." << endl;
	cerr << "  -a <N> <m_min> <m_max>  = fit all settings of a hyper-Erlang distribution and return the best fitted distribution." << endl;
	cerr << "                            <N> is the overall number of (internal) states used by the hyper-Erlang distribution," << endl;
	cerr << "                            <m_min> is the minimal number of Erlang branches to be used and" << endl;
	cerr << "                            <m_max> is the maximal number of Erlang branches (1 <= m_min <= m_max <= N)." << endl;
	cerr << "                            (default: N=" << DEFAULT_N << ", m_min=" << DEFAULT_M_MIN << ", m_max=" << DEFAULT_M_MAX << ")" << endl;
	cerr << "  -as <N> <m_min> <m_max> = apply the 'fit all'-algorithm in 3 rounds with priority queueing of best fitted distributions." << endl;
	cerr << "                            Parameters have the same meaning as in '-a' option." << endl;
	cerr << "  -c <N> <m_min> <m_max>  = calculate the overall number of settings. Parameters have the same meaning as in '-a' option." << endl;
	cerr << "  -e <eps>                = set convergence epsilon (default eps=" << DEFAULT_EPS << ")." << endl;
	cerr << "  -rf <gfitfile>          = read inital parameters from G-FIT output file <gfitfile>." << endl;
	cerr << "  -par                    = convergence check based on maximal difference of parameter values." << endl;
	cerr << "  -llh                    = convergence check based on relative difference of log-likelihood values (default)." << endl;
}


//---------------------------------------------------------------------
// Main Program of EM fitting algorithm for hyper-Erlang distributions
//---------------------------------------------------------------------
int main (int argc, char **argv)
{

	cout.precision(10);
	cout << "G-FIT V1.0" << endl;
	cout << "Fitting hyper-Erlang distributions with the EM-algorithm" << endl;
	cout << "Developed by Axel Thuemmler, July/August 2004" << endl << endl;

	if (argc < 2)
	{
		usage (argv);
		exit (-1);
	}

	unsigned int N = DEFAULT_N;
	unsigned int m_min = DEFAULT_M_MIN;
	unsigned int m_max = DEFAULT_M_MAX;

	// default: fit all hyper-Erlang distributions
	TFitAlgorithm type = FIT_ALL;

	// default: check convergence based on difference of log-likelihood values
	TConvCheck cvcheck = LLH_DIFF;

	// default values of convergence epsilon
	double eps = DEFAULT_EPS;

	// initialize default hyper-Erlang distribution
	TGerlangDist* gdist = new TGerlangDist(1);

	bool err = false;
	unsigned int opt_ind[8] = {0,0,0,0,0,0,0,0};

	// get input parameter
	if (argc > 2)
	{
		int pos = 2;
		while (pos < argc)
		{
			if (pos+1 < argc)
			{
				if (strcmp(argv[pos], "-f") == 0) opt_ind[0] = pos+1;
				if (strcmp(argv[pos], "-a") == 0) opt_ind[1] = pos+1;
				if (strcmp(argv[pos], "-as") == 0) opt_ind[2] = pos+1;
				if (strcmp(argv[pos], "-c") == 0) opt_ind[3] = pos+1;
				if (strcmp(argv[pos], "-e") == 0) opt_ind[4] = pos+1;
				if (strcmp(argv[pos], "-rf") == 0) opt_ind[5] = pos+1;
			}
			if (strcmp(argv[pos], "-par") == 0) opt_ind[6] = pos;
			if (strcmp(argv[pos], "-llh") == 0) opt_ind[7] = pos;
			pos++;
		}

		// fit GED option?
		if (opt_ind[0] > 0)
		{
			type = FIT;
			pos = opt_ind[0];
			int m = atol(argv[pos]);
			if (pos+m < argc)
			{
				unsigned int *r = new unsigned int[m];
				for (int i = 0; i < m; i++)
				{
					r[i] = atol(argv[pos+i+1]);
				}
					
				// initialize hyper-Erlang distribution with given parameters
				gdist->adjust_m(m);
				gdist->init(r);
		
				delete [] r;
			}
			else err = true;
		}

		// fit all option?
		if (opt_ind[1] > 0)
		{
			type = FIT_ALL;
			pos = opt_ind[1];
			if (pos+2 < argc)
			{
				N = atol(argv[pos]);
				m_min = atol(argv[pos+1]);
				m_max = atol(argv[pos+2]);
			}
			else err = true;
		}

		if (opt_ind[2] > 0)
		{
			type = FIT_ALL_STEPS;
			pos = opt_ind[2];
			if (pos+2 < argc)
			{
				N = atol(argv[pos]);
				m_min = atol(argv[pos+1]);
				m_max = atol(argv[pos+2]);
			}
			else err = true;
		}

		if (opt_ind[3] > 0)
		{
			type = CALC_NUM;
			pos = opt_ind[3];
			if (pos+2 < argc)
			{
				N = atol(argv[pos]);
				m_min = atol(argv[pos+1]);
				m_max = atol(argv[pos+2]);
			}
			else err = true;
		}
		
		if (opt_ind[4] > 0) eps = atof(argv[opt_ind[4]]);

		if (opt_ind[5] > 0)
		{
			type = FIT;
			gdist->read_file(argv[opt_ind[5]]);
		}

		if (opt_ind[6] > 0) cvcheck = PAR_DIFF;
		if (opt_ind[7] > 0) cvcheck = LLH_DIFF;

		if (err)
		{
			usage (argv);
			cerr << "ERROR: wrong number of arguments!!!" << endl;
			exit (-1);
		}
	}


	// read trace from file
	TTrace* trace = new TTrace;
	trace->read_file(argv[1]);
	trace->compute_distribution(PLOT_WIDTH, PLOT_STEPS);
	trace->compute_statistics();

#ifdef WRITE_SORTED_TRACE
	trace->sort();
	ostringstream filename;
	filename << argv[1] << "_sorted" << ends;
	trace->write_file(filename.str().c_str(), false);
#endif

	// initialize EM fitting algorithm for hyper-Erlang distributions
	TGerlangFit gfit(trace, eps, cvcheck);

	// get the number of seconds elapsed since 00:00 hours, Jan 1, 1970 UTC from the system clock
	time_t solutiontime = time(NULL);

	switch (type)
	{
		case CALC_NUM:
			cout << endl << "Number of settings = " << gfit.get_num_settings(N, m_min, m_max) << endl << endl;
			delete trace;
			delete gdist;
			return 0;
		case FIT:
			gfit.fit(gdist);
			break;
		case FIT_ALL:
			gfit.fit_all(gdist, N, m_min, m_max);
			break;
		case FIT_ALL_STEPS:
			gfit.fit_all_steps(gdist, N, m_min, m_max);
			break;
	}

	solutiontime = time(NULL) - solutiontime;

	cout << endl;
	cout << "iteration finished with accuracy " << gfit.get_max_diff() << endl;
	cout << "--> number of iterations:    " << gfit.get_count() << endl;
	cout << "--> elapsed CPU time in sec: " << solutiontime << endl;
	cout << "--> CPU time per iteration:  " << (double) solutiontime / (double) gfit.get_count() << endl;
	cout << endl;

	cout << "final r-vector:" << endl;
	gdist->print_r();
	cout << endl << "final lambda vector:" << endl;
	gdist->print_lambdas();
	cout << endl << "final starting probabilities:" << endl;
	gdist->print_weights();
	cout << endl;
	cout << "log-likelihood: " << gdist->compute_loglikelihood(trace) << endl;
	cout << endl;
	
	// compute statistics for fitted hyper-Erlang distribution
	gdist->compute_distribution(GFIT_FILENAME, PLOT_WIDTH, PLOT_STEPS);
	gdist->compute_statistics(trace->get_statistics());
	gdist->write_file("gfit.result");

	delete trace;
	delete gdist;

	return 0;
}
