/****************************************************************************/
/** G-FIT V1.0                                                             **/
/** Fitting hyper-Erlang distributions to trace data with the EM algorithm **/
/**                                                                        **/
/** Developed by Axel Thuemmler, July/August 2004                          **/
/** Copyright (c) 2004 University of Dortmund. All rights reserved.        **/
/**                                                                        **/
/** The G-FIT software package is distributed in open-source under the     **/
/** terms of the GNU general public license agreement. To document the     **/
/** usage of G-FIT, we ask each user to cite our scientific publications   **/
/** mentioned on our Web site.                                             **/
/****************************************************************************/
//------------------------------------------------------------------
// Implementation of priority queue with hyper-Erlang distributions
//------------------------------------------------------------------

#include <iostream>
#include <cfloat>
#include <cstdlib>

#include "queue.h"

using namespace std;


//-------------
// Constructor
//-------------
TQueue::TQueue(unsigned int _size, unsigned int m_max)
{
	size = _size;
	queue = new TGED*[size];

	for(unsigned int i = 0; i < size; i++)
	{
		queue[i] = new TGED;
		queue[i]->loglikelihood = -DBL_MAX;
		queue[i]->r = new unsigned int[m_max];
		queue[i]->lambdas = new double[m_max];
		queue[i]->weights = new double[m_max];
	}

	// queue is empty at the beginning
	num_in = 0;
}


//------------
// Destructor
//------------
TQueue::~TQueue()
{

	for(unsigned int i = 0; i < size; i++)
	{
		delete [] queue[i]->r;
		delete [] queue[i]->lambdas;
		delete [] queue[i]->weights;
		delete queue[i];
	}

	delete [] queue;
}


//-----------------------------------------------
// Insert new GED sorted into the priority queue
//-----------------------------------------------
void TQueue::insert(TGED ged)
{
	unsigned int pos = num_in;

	// queue full?
	if (num_in == size)
	{
		// should GED be inserted into the priority queue?
		if (ged.loglikelihood > queue[size-1]->loglikelihood)
		{
			pos--;
		}
	}
	else
	{
		num_in++;
	}

	if (pos < size)
	{
		insert_pos(pos, ged);
		
		TGED* swap;

		while ((pos > 0) && (queue[pos]->loglikelihood > queue[pos-1]->loglikelihood))
		{
			swap = queue[pos-1];
			queue[pos-1] = queue[pos];
			queue[pos] = swap;
			pos--;
		}
	}
}


//------------------------------------------------
// Update the GED in the queue at position pos
// and move it to a better position if neccessary
// to remain the priority queue in sorted order
//------------------------------------------------ 
void TQueue::update(unsigned int pos, double llh, double* lambdas, double* weights)
{
	TGED* swap;

	queue[pos]->loglikelihood = llh;
	for(unsigned int i = 0; i < queue[pos]->m; i++)
	{
		queue[pos]->lambdas[i] = lambdas[i];
		queue[pos]->weights[i] = weights[i];
	}

	// get a better position in the priority queue?
	while ((pos > 0) && (queue[pos]->loglikelihood > queue[pos-1]->loglikelihood))
	{
		swap = queue[pos-1];
		queue[pos-1] = queue[pos];
		queue[pos] = swap;
		pos--;
		cout << "improving position in priority queue!!!" << endl;
	}

	// get a worse position in the priority queue?
	// this should not be possible since the likelihood values
	// are increased with every iteration of the EM-algorithm
	while ((pos < size-1) && (queue[pos]->loglikelihood < queue[pos+1]->loglikelihood))
	{
		swap = queue[pos+1];
		queue[pos+1] = queue[pos];
		queue[pos] = swap;
		pos++;
		cout << "ATTENTION: POSITION GETS WORSE !!!!!!" << endl;
	}
}


//----------------------------
// Return GED at position pos
//----------------------------
void TQueue::get(unsigned int pos, TGerlangDist* gdist)
{
	gdist->adjust_m(queue[pos]->m);
	gdist->init(queue[pos]->r, queue[pos]->lambdas, queue[pos]->weights);
}


//----------------------------------------------------------
// Return best GED with respect to the log-likelihood value
//----------------------------------------------------------
void TQueue::get_bestfit(TGerlangDist* gdist)
{
	get(0, gdist);
}


unsigned int TQueue::get_size()
{
	return num_in;
}


//--------------------------------------------------------------
// Print num GEDs of the priority queue starting with the first
//--------------------------------------------------------------
void TQueue::print(unsigned int num)
{
	if (num > num_in) num = num_in;

	for(unsigned int i = 0; i < num; i++)
	{
		cout << "element no. " << i+1 << endl;
		cout << "  r-vector: ";
		for (unsigned int k = 0; k < queue[i]->m-1; k++) cout << queue[i]->r[k] << ", ";
		cout << queue[i]->r[queue[i]->m-1] << endl;
		cout << "  log-likelihood = " << queue[i]->loglikelihood << endl;
	}
}



//****************** PRIVATE METHODS **********************************


//--------------------------------
// Insert GED at a given position
//--------------------------------
void TQueue::insert_pos(unsigned int pos, TGED ged)
{
	queue[pos]->loglikelihood = ged.loglikelihood;
	queue[pos]->m = ged.m;
	for(unsigned int i = 0; i < ged.m; i++)
	{
		queue[pos]->r[i] = ged.r[i];
		queue[pos]->lambdas[i] = ged.lambdas[i];
		queue[pos]->weights[i] = ged.weights[i];
	}
}
