/****************************************************************************/
/** G-FIT V1.0                                                             **/
/** Fitting hyper-Erlang distributions to trace data with the EM algorithm **/
/**                                                                        **/
/** Developed by Axel Thuemmler, July/August 2004                          **/
/** Copyright (c) 2004 University of Dortmund. All rights reserved.        **/
/**                                                                        **/
/** The G-FIT software package is distributed in open-source under the     **/
/** terms of the GNU general public license agreement. To document the     **/
/** usage of G-FIT, we ask each user to cite our scientific publications   **/
/** mentioned on our Web site.                                             **/
/****************************************************************************/
//--------------------------------
// Implementation of class TTrace
//--------------------------------

#include <iostream>
#include <fstream>
#include <cmath>
#include <cfloat>
#include <cstdlib>

#include "trace.h"

using namespace std;

//-------------
// Constructor
//-------------
TTrace::TTrace()
{
	trace = NULL;
	size = 0;
}

//---------------
// Constructor 2
//---------------
TTrace::TTrace(double* _trace, unsigned long _size)
{
	init(_trace, _size);
}


void TTrace::init(double* _trace, unsigned long _size)
{
	size = _size;
	trace = new double[size];
	for (unsigned long i = 0; i < size; i++)
	{
		trace[i] = _trace[i];
	}
}



//------------
// Destructor
//------------
TTrace::~TTrace()
{
	delete [] trace;
}


unsigned long TTrace::get_size()
{
	return size;
}

double TTrace::get_element(unsigned int i)
{
	return trace[i];
}

TTraceStatistics TTrace::get_statistics()
{
	return statistics;
}


//----------------------
// Read trace from file
//----------------------
void TTrace::read_file(const char* filename)
{

	cerr << "reading trace from file ...";

	ifstream tracefile(filename);
	if (!tracefile)
	{
		cerr << "ERROR: trace file not found !!!" << endl;
		exit (-1);
	}

	char line[257];
	bool first_line = true;

	unsigned long pos = 0;

	while (tracefile.getline(line, 256))
	{
		if (strcmp(line, "") == 0) continue;	// row empty => next row
		if (line[0] == '#') continue;			// row with comment => next row

		if (first_line)							// first row contains size of trace
		{
			size = atol(line);
			trace = new double[size];
			first_line = false;
		}
		else
		{
			trace[pos] = atof(line);
			pos++;
		}
	}

	tracefile.close();

	cout << endl;	
}


//---------------------
// Write trace to file
//---------------------
void TTrace::write_file(const char* filename, bool with_size)
{
	cerr << "writing trace to file ...";

	ofstream tracefile(filename);

	if (with_size) tracefile << size << endl;

	for (unsigned long n = 0; n < size; n++)
	{
		tracefile << trace[n] << endl;
	}

	tracefile.close();
	cout << endl;
}


int compare_double(const void* a, const void* b)
{
	double result = *(double*)a - *(double*)b;

	if (result == 0) return 0;
	if (result < 0) return (-1);
	
	return 1;
}

double TTrace::sort()
{
	cerr << "sorting trace ...";

	qsort(trace, size, sizeof(double), compare_double);

	cout << endl;

	double median;

	if (size % 2 == 0) median = double (trace[size/2] + trace[size/2 + 1]) / 2.0;
	else median = trace[(size+1)/2];

	return median;
}


void TTrace::scale(double mean)
{
	double m = 0.0;
	for (unsigned long i = 0; i < size; i++)
	{
		m = (double) i / (double) (i+1) * m + 1.0 / (double) (i+1) * trace[i];
	}

	for (unsigned long i = 0; i < size; i++)
	{
		trace[i] = mean * trace[i] / m;
	}
}


//------------------------------------------------------------
// Compute estimates for the pdf, cdf, and ccdf of empirical
// distribution from given trace data
//   width = width of histogram 
//   steps = steps of histogram
//           (histogram goes from 0 to width*steps)
//------------------------------------------------------------
void TTrace::compute_distribution(double width, unsigned int steps)
{

	cerr << "computing trace distribution ...";

	unsigned int *histogram = new unsigned int[steps+1];

	//initialize histogram with zero
	for (unsigned int i = 0; i < steps+1; i++) histogram[i] = 0;
	
	unsigned int j;
	for (unsigned long n = 0; n < size; n++)
	{
		j = (unsigned int) (trace[n] / width);
		if (j < steps)
		{
			histogram[j]++;
		}
		else
		{
			histogram[steps]++;
		}
	}
	
	// Write pdf, cdf and ccdf of empirical distribution
	ofstream t_pdf_file(TRACE_PDF_FILENAME);
	ofstream t_cdf_file(TRACE_CDF_FILENAME);
	ofstream t_ccdf_file(TRACE_CCDF_FILENAME);

	double sum = 0.0;
	t_cdf_file << "0.0  0.0" << endl;
	t_ccdf_file << "0.0  1.0" << endl;

	for (unsigned int n = 0; n < steps; n++)
	{
		t_pdf_file << n * width << "  " << (double) histogram[n] / size / width << endl;
		t_pdf_file << (n+1) * width << "  " << (double) histogram[n] / size / width << endl;

		sum += (double) histogram[n] / size;
		t_cdf_file << (n+1) * width << "  " << sum << endl;
		t_ccdf_file << (n+1) * width << "  " << 1.0 - sum << endl;
	}

	t_pdf_file.close();
	t_cdf_file.close();
	t_ccdf_file.close();

	delete [] histogram;

}


//----------------------------------------------------------------------
// Compute estimates for first five moments, variance, coefficient of
// variation, minimum, maximum for the given trace data
//----------------------------------------------------------------------
void TTrace::compute_statistics()
{

	// computation of moments, etc.
	for (unsigned int i = 0; i < 5; i++) statistics.m[i] = 0.0;
	statistics.var = 0.0;
	statistics.min = DBL_MAX;
	statistics.max = 0.0;

	// compute estimates for the first five moments 
	for (unsigned long i = 0; i < size; i++)
	{
		statistics.m[0] = (double) i / (double) (i+1) * statistics.m[0] + 1.0 / (double) (i+1) * trace[i];
		statistics.m[1] = (double) i / (double) (i+1) * statistics.m[1] + 1.0 / (double) (i+1) * pow(trace[i],2);
		statistics.m[2] = (double) i / (double) (i+1) * statistics.m[2] + 1.0 / (double) (i+1) * pow(trace[i],3);
		statistics.m[3] = (double) i / (double) (i+1) * statistics.m[3] + 1.0 / (double) (i+1) * pow(trace[i],4);
		statistics.m[4] = (double) i / (double) (i+1) * statistics.m[4] + 1.0 / (double) (i+1) * pow(trace[i],5);

		if (statistics.min > trace[i]) statistics.min = trace[i];
		if (statistics.max < trace[i]) statistics.max = trace[i]; 
	}

	// compute estimate for the variance and coefficient of variation
	for (unsigned long i = 0; i < size; i++)
	{
		statistics.var += (trace[i] - statistics.m[0]) * (trace[i] - statistics.m[0]);
	}
	statistics.var /= (double) (size - 1);
	statistics.cvs = statistics.var / (statistics.m[0] * statistics.m[0]);

	cout << endl;

	cout << "First   Moment         : " << statistics.m[0] << endl;
	cout << "Second  Moment         : " << statistics.m[1] << endl;
	cout << "Third   Moment         : " << statistics.m[2] << endl;
//	cout << "Fourth  Moment         : " << statistics.m[3] << endl;
//	cout << "Fifth   Moment         : " << statistics.m[4] << endl;
	cout << "Variance               : " << statistics.var << endl;
	cout << "Squared Coeff. of Var. : " << statistics.cvs << endl;
	cout << "Minimum                : " << statistics.min << endl;
	cout << "Maximum                : " << statistics.max << endl;

}
